
/**
 * AppPreferences -- Preferences for the Plot Digitizer application.
 *
 * Copyright (C) 2000-2023, Joseph A. Huwaldt. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite
 * 330, Boston, MA 02111-1307, USA. Or visit: http://www.gnu.org/licenses/gpl.html
 */

import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.EscapeJDialog;
import jahuwaldt.swing.MDIApplication;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeSupport;
import java.text.NumberFormat;
import java.util.List;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.Preferences;
import javax.swing.*;

/**
 * This class serves as a collection of preferences for the Plot Digitizer application.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt, Date: May 2, 2004
 * @version January 1, 2024
 */
public class AppPreferences implements jahuwaldt.swing.Preferences {

    private final Preferences prefs = Preferences.userNodeForPackage(this.getClass());
	
    //  Use a change-listener system to keep the preferences dialog updated
    private final PropertyChangeSupport propertyChange = new PropertyChangeSupport(this);

    // Defines the shape of the point symbol
    public enum SymbolShape {
        CROSS, CIRCLE
    }

    //  Preference keys
    private static final String LASTPATH = "lastPath";
    private static final String XMIN_FOR_YMIN = "xMinForYMin";
    private static final String XAXIS_COLOR = "FBXAxisColor";
    private static final String YAXIS_COLOR = "FBYAxisColor";
    private static final String SYMBOL_COLOR = "FBSymbolColor";
    private static final String LINE_COLOR = "FBLineColor";
    private static final String SYMBOL_SIZE = "FBSymbolSize";
    private static final String SYMBOL_SHAPE = "FBSymbolShape";
    private static final String USE_LINES = "FBUseLines";
    private static final String OUTPUT_PIXELS = "OutputPixels";
    private static final String NUM_GRIDLINES = "numGridLines";
    private static final String GRIDLINE_COLOR = "GridLineColor";
    
    //  Default preference values.
    private static final boolean DEF_XMIN_FOR_YMIN = false;
    private static final Color DEF_XAXIS_COLOR = Color.RED;
    private static final Color DEF_YAXIS_COLOR = Color.BLUE;
    private static final Color DEF_SYMBOL_COLOR = Color.ORANGE;
    private static final Color DEF_LINE_COLOR = Color.YELLOW;
    private static final int DEF_SYMBOL_SIZE = 2;
    private static final SymbolShape DEF_SYMBOL_SHAPE = SymbolShape.CROSS;
    private static final boolean DEF_USE_LINES = true;
    private static final boolean DEF_OUTPUT_PIXELS = false;
    private static final int DEF_NUM_GRIDLINES = 10;
    private static final Color DEF_GRIDLINE_COLOR = Color.YELLOW;
    
    
    //  The resource bundle for this application.
    private final ResourceBundle RESOURCES = MDIApplication.getInstance().getResourceBundle();

     //  A reference to the window containing the applicatin preferences.
    private PrefsDialog prefsDialog = null;


    /**
     * Returns the file path to the parent of the last referenced file. Returns
     * <code>null</code> if no last path could be found.
     */
    @Override
    public String getLastPath() {
        return prefs.get(LASTPATH, null);
    }

    /**
     * Set the last file path referenced by the user. This is the path to the last parent
     * of the last referenced file.
     */
    @Override
    public void setLastPath(String path) {
		String oldPath = getLastPath();
		if ( (nonNull(path) && !path.equals(oldPath)) || (isNull(path) && nonNull(oldPath)) ) {
			prefs.put(LASTPATH, path);
            propertyChange.firePropertyChange(LASTPATH, oldPath, path);
        }
    }

    public boolean getXMinForYMin() {
        return prefs.getBoolean(XMIN_FOR_YMIN, DEF_XMIN_FOR_YMIN);
    }

    public void setXMinForYMin(boolean bool) {
        if (bool != getXMinForYMin())
            prefs.putBoolean(XMIN_FOR_YMIN, bool);
    }

    /**
     * Returns the color of the feed back X-axis.
     *
     * @return Color
     */
    public Color getXAxisColor() {
        return getColorPrefs(XAXIS_COLOR, DEF_XAXIS_COLOR);
    }

    /**
     * Returns the color of the feed back Y-axis.
     *
     * @return Color
     */
    public Color getYAxisColor() {
        return getColorPrefs(YAXIS_COLOR, DEF_YAXIS_COLOR);
    }

    /**
     * Returns the color of the feed back symbols.
     *
     * @return Color
     */
    public Color getSymbolColor() {
        return getColorPrefs(SYMBOL_COLOR, DEF_SYMBOL_COLOR);
    }

    /**
     * Returns the color of the feed back symbols.
     *
     * @return Color
     */
    public Color getLineColor() {
        return getColorPrefs(LINE_COLOR, DEF_LINE_COLOR);
    }

    /**
     * Returns the size of the feed back symbols.
     *
     * @return Size of the Symbol
     */
    public int getSymbolSize() {
        return prefs.getInt(SYMBOL_SIZE, DEF_SYMBOL_SIZE);
    }

    /**
     * Returns the type of symbol shape selected
     *
     * @return SymbolShape, an enum defining the shape of the symbol
     */
    public SymbolShape getSymbolShape() {
        return SymbolShape.valueOf(prefs.get(SYMBOL_SHAPE, DEF_SYMBOL_SHAPE.toString()));
    }

    /**
     * Returns true if lines should be drawn between digitized points, false otherwise.
     *
     * @return boolean
     */
    public boolean getUseLines() {
        return prefs.getBoolean(USE_LINES, DEF_USE_LINES);
    }

    /**
     * Returns true if the program should output untransformed pixels in addition to the
     * transformed values.
     *
     * @return boolean
     */
    public boolean getOutputPixels() {
        return prefs.getBoolean(OUTPUT_PIXELS, DEF_OUTPUT_PIXELS);
    }

    public int getGridIncrement() {
        return prefs.getInt(NUM_GRIDLINES, DEF_NUM_GRIDLINES);
    }

    public Color getGridLineColor() {
        return getColorPrefs(GRIDLINE_COLOR, DEF_GRIDLINE_COLOR);
    }

    /**
     * Return the preference with the specified key String.
     *
     * @param key The key String identifying the preference to be retrieved.
     */
    @Override
    public String get(String key) {
		return prefs.get(key, null);
    }

    /**
     * Set the preference with the specified key String.
     *
     * @param key   The key String identifying the preference to be set.
     * @param value THe String value to store as the preference.
     */
    @Override
    public void set(String key, String value) {
		prefs.put(key, value);
    }

    /**
     * Method that displays a dialog that allows the user to change the application
     * preferences.
     */
    @Override
    public void showPreferenceDialog() {
        if (prefsDialog != null)
            prefsDialog.setVisible(true);

        else {
            prefsDialog = new PrefsDialog();
            prefsDialog.pack();
            AppUtilities.positionWindow(prefsDialog, prefsDialog.getWidth(), prefsDialog.getHeight());
            prefsDialog.setVisible(true);
        }
    }

    /**
     * Method that parses a color preference from the Preferences and returns a Color object.
     */
    private Color getColorPrefs(String key, Color defColor) {
        int red = prefs.getInt(key + "R", defColor.getRed());
        int green = prefs.getInt(key + "G", defColor.getGreen());
        int blue = prefs.getInt(key + "B", defColor.getBlue());
        int alpha = prefs.getInt(key + "A", defColor.getAlpha());

        Color color = new Color(red, green, blue, alpha);
        return color;
    }

    /**
     * Method that encodes a color preference into the Preferences.
     */
    private void encodeColorPrefs(String key, Color color) {
        prefs.putInt(key + "R", color.getRed());
        prefs.putInt(key + "G", color.getGreen());
        prefs.putInt(key + "B", color.getBlue());
        prefs.putInt(key + "A", color.getAlpha());
    }

    /**
     * Method that creates a default set of preferences for use in this program.
     *
     * @param props Properties object to be set
     * @return Properties object with members set
     */
    private void setDefaultPreferences() {
        prefs.putBoolean(XMIN_FOR_YMIN, DEF_XMIN_FOR_YMIN);
        
        encodeColorPrefs(XAXIS_COLOR, DEF_XAXIS_COLOR);
        encodeColorPrefs(YAXIS_COLOR, DEF_YAXIS_COLOR);
        encodeColorPrefs(SYMBOL_COLOR, DEF_SYMBOL_COLOR);
        encodeColorPrefs(LINE_COLOR, DEF_LINE_COLOR);
        
        prefs.putInt(SYMBOL_SIZE, DEF_SYMBOL_SIZE);
        prefs.put(SYMBOL_SHAPE, DEF_SYMBOL_SHAPE.toString());
        prefs.putBoolean(USE_LINES, DEF_USE_LINES);
        
        prefs.putBoolean(OUTPUT_PIXELS, DEF_OUTPUT_PIXELS);
        
        prefs.putInt(NUM_GRIDLINES, DEF_NUM_GRIDLINES);
        encodeColorPrefs(GRIDLINE_COLOR, DEF_GRIDLINE_COLOR);

    }

    /**
     * Handles the user changing the symbol size
     *
     * @param e
     */
    public void handleSymbolSize(ActionEvent e) {
        JComboBox cb = (JComboBox) e.getSource();
        int symbolSize = Integer.valueOf((String) cb.getSelectedItem());
        if (symbolSize != getSymbolSize()) {
            prefs.putInt(SYMBOL_SIZE, symbolSize);
            updateDisplay();
        }
    }

    public void handleNumGridLines(ActionEvent e) {
        JComboBox cb = (JComboBox) e.getSource();
        int numGridLines = Integer.valueOf((String) cb.getSelectedItem());
        if (numGridLines != getGridIncrement()) {
            prefs.putInt(NUM_GRIDLINES, numGridLines);
            updateDisplay();
        }
    }

    public void handleGridColor(ActionEvent e) {
        Color origColor = getGridLineColor();
        String msg = RESOURCES.getString("chooseGridLineColor");
        Color newColor = JColorChooser.showDialog(prefsDialog, msg, origColor);
        if (nonNull(newColor) && !newColor.equals(origColor)) {
            //  The user has changed the color, make the appropriate preference changes.
            encodeColorPrefs(GRIDLINE_COLOR, newColor);

            //  Repaint the color swatch in the dialog.
            colorSwatch(prefsDialog.gridColorImg, newColor);
            
            //  Update the display.
            updateDisplay();
        }
    }

    /**
     * Handles setting the preference for the symbol shape
     *
     * @param e The event that caused this method to be called.
     */
    public void handleSymbolShape(ActionEvent e) {
        JComboBox cb = (JComboBox) e.getSource();
        String format = (String) cb.getSelectedItem();
        if (format.toUpperCase().contains("CROSS")) {
            prefs.put(SYMBOL_SHAPE, SymbolShape.CROSS.toString());
        } else if (format.contains("CIRCLE")) {
            prefs.put(SYMBOL_SHAPE, SymbolShape.CIRCLE.toString());
        }
        updateDisplay();
    }

    /**
     * Method that handles the user clicking on the X-Axis Color Swatch.
     */
    public void handleXAxisColor(ActionEvent e) {
        Color origColor = getXAxisColor();
        String msg = RESOURCES.getString("chooseXAxisColor");
        Color newColor = JColorChooser.showDialog(prefsDialog, msg, origColor);
        if (nonNull(newColor) && !newColor.equals(origColor)) {
            //  The user has changed the color, make the appropriate preference changes.
            encodeColorPrefs(XAXIS_COLOR, newColor);

            //  Repaint the color swatch in the dialog.
            colorSwatch(prefsDialog.xAxisColorImg, newColor);

            //  Update the display.
            updateDisplay();
        }
    }

    /**
     * Method that handles the user clicking on the Y-Axis Color Swatch.
     */
    public void handleYAxisColor(ActionEvent e) {
        Color origColor = getYAxisColor();
        String msg = RESOURCES.getString("chooseYAxisColor");
        Color newColor = JColorChooser.showDialog(prefsDialog, msg, origColor);
        if (nonNull(newColor) && !newColor.equals(origColor)) {
            //  The user has changed the color, make the appropriate preference changes.
            encodeColorPrefs(YAXIS_COLOR, newColor);

            //  Repaint the color swatch in the dialog.
            colorSwatch(prefsDialog.yAxisColorImg, newColor);

            //  Update the display.
            updateDisplay();
        }
    }

    /**
     * Method that handles the user clicking on the Symbol Color Swatch.
     */
    public void handleSymbolColor(ActionEvent e) {
        Color origColor = getSymbolColor();
        String msg = RESOURCES.getString("chooseSymbolColor");
        Color newColor = JColorChooser.showDialog(prefsDialog, msg, origColor);
        if (nonNull(newColor) && !newColor.equals(origColor)) {
            //  The user has changed the color, make the appropriate preference changes.
            encodeColorPrefs(SYMBOL_COLOR, newColor);

            //  Repaint the color swatch in the dialog.
            colorSwatch(prefsDialog.symbolColorImg, newColor);

            //  Update the display.
            updateDisplay();
        }
    }

    /**
     * Method that handles the user clicking on the Line Color Swatch.
     */
    public void handleLineColor(ActionEvent e) {
        Color origColor = getLineColor();
        String msg = RESOURCES.getString("chooseLineColor");
        Color newColor = JColorChooser.showDialog(prefsDialog, msg, origColor);
        if (nonNull(newColor) && !newColor.equals(origColor)) {
            //  The user has changed the color, make the appropriate preference changes.
            encodeColorPrefs(LINE_COLOR, newColor);

            //  Repaint the color swatch in the dialog.
            colorSwatch(prefsDialog.lineColorImg, newColor);

            //  Update the display.
            updateDisplay();
        }
    }

    /**
     * Method that handles the user clicking on the Restore Defaults button.
     */
    public void handleSetDefaults(ActionEvent e) {
        //  Reset the default property values.
        setDefaultPreferences();
 
        //  Repaint all the color swatches.
        colorSwatch(prefsDialog.xAxisColorImg, getXAxisColor());
        colorSwatch(prefsDialog.yAxisColorImg, getYAxisColor());
        colorSwatch(prefsDialog.symbolColorImg, getSymbolColor());
        colorSwatch(prefsDialog.lineColorImg, getLineColor());
        colorSwatch(prefsDialog.gridColorImg, getGridLineColor());

        //  Redraw the checkboxes if necessary.
        prefsDialog.useLinesBox.setSelected(getUseLines());
        prefsDialog.pixelOutputBox.setSelected(getOutputPixels());

        //  Update the display.
        updateDisplay();
    }

    /**
     * Method that handles the user clicking on the "use lines" check box.
     */
    public void handleUseLines(ActionEvent e) {
        JCheckBox check = (JCheckBox) e.getSource();
        if (check.isSelected()) {
            prefs.putBoolean(USE_LINES, true);
        } else {
            prefs.putBoolean(USE_LINES, false);
        }

        //  Update the display.
        updateDisplay();
    }

    /**
     * Method that handles the user clicking on the "output pixels" check box.
     */
    public void handleOutputPixels(ActionEvent e) {
        JCheckBox check = (JCheckBox) e.getSource();
        if (check.isSelected()) {
            prefs.putBoolean(OUTPUT_PIXELS, true);
        } else {
            prefs.putBoolean(OUTPUT_PIXELS, false);
        }

        //  Update the display.
        updateDisplay();
    }

    /**
     * Repaint all objects affected by a color change.
     */
    private void updateDisplay() {
        //  Repaint the preferences dialog.
        prefsDialog.repaint();

        //  Repaint all the open windows to make the change take effect.
        List windows = MDIApplication.allOpenWindows();
        int size = windows.size();
        for (int i = 0; i < size; ++i) {
            Frame window = (Frame) windows.get(i);
            window.repaint();
        }
    }

    /**
     * Method that colors in an image with the specified color.
     */
    private static void colorSwatch(BufferedImage image, Color color) {
        Graphics gr = image.getGraphics();
        gr.setColor(color);
        gr.fillRect(0, 0, image.getWidth(), image.getHeight());
    }

    /**
     * The model dialog that allows the user to set this application's preferences.
     */
    private class PrefsDialog extends EscapeJDialog {

        private final int kImgWidth = 128;
        private final int kImgHeight = 32;
        BufferedImage xAxisColorImg = new BufferedImage(kImgWidth, kImgHeight, BufferedImage.TYPE_INT_ARGB);
        BufferedImage yAxisColorImg = new BufferedImage(kImgWidth, kImgHeight, BufferedImage.TYPE_INT_ARGB);
        BufferedImage symbolColorImg = new BufferedImage(kImgWidth, kImgHeight, BufferedImage.TYPE_INT_ARGB);
        BufferedImage lineColorImg = new BufferedImage(kImgWidth, kImgHeight, BufferedImage.TYPE_INT_ARGB);
        BufferedImage gridColorImg = new BufferedImage(kImgWidth, kImgHeight, BufferedImage.TYPE_INT_ARGB);

        JCheckBox useLinesBox, pixelOutputBox;

        /**
         * Construct a new preferences dialog window.
         */
        @SuppressWarnings("OverridableMethodCallInConstructor")
        PrefsDialog() {
            super(null, RESOURCES.getString("prefsPanelTitle"), false);
            ResourceBundle resBundle = RESOURCES;

            //  Layout the Line Settings window.
            JPanel cp = new JPanel();
            cp.setLayout(new BorderLayout());

            //	Add instructions at the top.
            JPanel panel = new JPanel();
            panel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createLineBorder(Color.black),
                    resBundle.getString("lineSymbolColorsTitle")));
            cp.add(panel, BorderLayout.NORTH);

            //  Color each of the swatches.
            colorSwatch(xAxisColorImg, getXAxisColor());
            colorSwatch(yAxisColorImg, getYAxisColor());
            colorSwatch(symbolColorImg, getSymbolColor());
            colorSwatch(lineColorImg, getLineColor());
            colorSwatch(gridColorImg, getGridLineColor());

            //  Layout the labels for the color swatches.
            Box box = Box.createHorizontalBox();
            panel.add(box);
            panel = new JPanel(new GridLayout(0, 1));
            panel.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
            box.add(panel);
            panel.add(new JLabel(resBundle.getString("xAxisColorLabel")));
            panel.add(new JLabel(resBundle.getString("yAxisColorLabel")));
            panel.add(new JLabel(resBundle.getString("symbolColorLabel")));
            panel.add(new JLabel(resBundle.getString("lineColorLabel")));
            panel.add(new JLabel(resBundle.getString("gridColorLabel")));

            try {
                //  Lay out the color swatches.
                panel = new JPanel(new GridLayout(0, 1));
                box.add(panel);
                JButton button = new JButton(new ImageIcon(xAxisColorImg));
                button.setBorder(BorderFactory.createEmptyBorder(2, 0, 2, 0));
                button.addActionListener(AppUtilities.getActionListenerForMethod(AppPreferences.this, "handleXAxisColor"));
                panel.add(button);

                button = new JButton(new ImageIcon(yAxisColorImg));
                button.setBorder(BorderFactory.createEmptyBorder(2, 0, 2, 0));
                button.addActionListener(AppUtilities.getActionListenerForMethod(AppPreferences.this, "handleYAxisColor"));
                panel.add(button);

                button = new JButton(new ImageIcon(symbolColorImg));
                button.setBorder(BorderFactory.createEmptyBorder(2, 0, 2, 0));
                button.addActionListener(AppUtilities.getActionListenerForMethod(AppPreferences.this, "handleSymbolColor"));
                panel.add(button);

                button = new JButton(new ImageIcon(lineColorImg));
                button.setBorder(BorderFactory.createEmptyBorder(2, 0, 2, 0));
                button.addActionListener(AppUtilities.getActionListenerForMethod(AppPreferences.this, "handleLineColor"));
                panel.add(button);

                button = new JButton(new ImageIcon(gridColorImg));
                button.setBorder(BorderFactory.createEmptyBorder(2, 0, 2, 0));
                button.addActionListener(AppUtilities.getActionListenerForMethod(AppPreferences.this, "handleGridColor"));
                panel.add(button);

                // create bottom panel
                panel = new JPanel(new GridLayout(0, 1));
                panel.setBorder(BorderFactory.createEmptyBorder(6, 6, 6, 6));
                cp.add("Center", panel);

                // Set up combo box for number of grid lines
                JLabel gridLabel = new JLabel(resBundle.getString("numGridLinesLabel"));
                panel.add(gridLabel);
                int numLines = Integer.parseInt(resBundle.getString("numGridLines"));
                String[] intString = new String[numLines];
                for (int i = 1; i <= numLines; ++i) {
                    intString[i - 1] = String.format("%d", i);
                }
                JComboBox gridLineOptions = new JComboBox(intString);
                // select current number of gridlines
                for (int i = 0; i < intString.length; ++i) {
                    if (((String) gridLineOptions.getItemAt(i)).equals(String.valueOf(getGridIncrement()))) {
                        gridLineOptions.setSelectedIndex(i);
                    }
                }

                // define action listener to handle the combobox
                gridLineOptions.addActionListener(AppUtilities.getActionListenerForMethod(AppPreferences.this, "handleNumGridLines"));

                // add combobox to panel
                panel.add(gridLineOptions);

                //  Create a check-box for using lines.
                useLinesBox = new JCheckBox(resBundle.getString("drawLinesCBLabel"), getUseLines());
                useLinesBox.addActionListener(AppUtilities.getActionListenerForMethod(AppPreferences.this, "handleUseLines"));
                panel.add(useLinesBox);

                //  Create a check-box for outputting untransformed pixels.
                pixelOutputBox = new JCheckBox(resBundle.getString("outputPixelsCBLabel"), getOutputPixels());
                pixelOutputBox.addActionListener(AppUtilities.getActionListenerForMethod(AppPreferences.this, "handleOutputPixels"));
                panel.add(pixelOutputBox);

                //  Create a "restore defaults" button.
                box = Box.createHorizontalBox();
                cp.add(box, BorderLayout.SOUTH);

                button = new JButton(resBundle.getString("defaultPrefsBtnLabel"));
                button.addActionListener(AppUtilities.getActionListenerForMethod(AppPreferences.this, "handleSetDefaults"));
                box.add(button);
                box.add(Box.createGlue());

                // set up Line setting tab
                JTabbedPane settingTabs = new JTabbedPane();

                // add Line settings to tab (all the UI code above)0
                settingTabs.addTab(resBundle.getString("lineSettingsTabLabel"), cp);

                // Set up Symbol settings such as size and shape tab
                JPanel symbolSettings = new JPanel();

                // setup symbol shape label
                JLabel shapeLabel = new JLabel(resBundle.getString("shapeSymbolLabel"));
                // setup combo box for symbol shape
                String[] shapeTypes = resBundle.getString("ShapeTypes").split(",");
                JComboBox symbolShapeType = new JComboBox(shapeTypes);
                for (int i = 0; i < shapeTypes.length; ++i) {
                    if (((String) symbolShapeType.getItemAt(i)).toUpperCase().contains(getSymbolShape().toString())) {
                        symbolShapeType.setSelectedIndex(i);
                    }
                }

                symbolShapeType.addActionListener(AppUtilities.getActionListenerForMethod(AppPreferences.this, "handleSymbolShape"));
                symbolSettings.add(shapeLabel);
                symbolSettings.add(symbolShapeType);

                // Setup symbol size label
                JLabel sizeLabel = new JLabel(resBundle.getString("shapeSymbolSizeLabel"));
                // Set up symbol size box
                NumberFormat fmt = NumberFormat.getIntegerInstance();
                String[] sizes = new String[10];
                for (int i = 0; i < 10; ++i) {
                    sizes[i] = fmt.format(i + 1);
                }
                JComboBox symbolSize = new JComboBox(sizes);
                for (int i = 0; i < sizes.length; ++i) {
                    if (((String) symbolSize.getItemAt(i)).toUpperCase().equals(getSymbolSize())) {
                        symbolSize.setSelectedIndex(i);
                    }
                }
                symbolSize.addActionListener(AppUtilities.getActionListenerForMethod(AppPreferences.this, "handleSymbolSize"));
                symbolSettings.add(sizeLabel);
                symbolSettings.add(symbolSize);

                // add symbol settings to tab
                settingTabs.addTab(resBundle.getString("symbolSettingsTabLabel"), symbolSettings);

                // add all tabs to window
                this.getContentPane().add(settingTabs);

            } catch (NoSuchMethodException e) {
                Logger logger = PlotDigitizer.getLogger();
                logger.log(Level.SEVERE, "Error creating PrefsDialog window", e);
                e.printStackTrace();
                AppUtilities.showException(null, resBundle.getString("prefErrTitle"), resBundle.getString("unexpectedMsg"), e);
            }

        }

    }

}
