/**
 *  PlotDigitizer  -- An application that digitizes points off an image.
 *
 *  Copyright (C) 2000-2023, Joseph A. Huwaldt.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */
import jahuwaldt.io.ExtFilenameFilter;
import jahuwaldt.swing.*;
import java.awt.FileDialog;
import java.awt.Frame;
import java.awt.Window;
import java.awt.desktop.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.*;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import static java.util.Objects.isNull;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.*;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;


/**
 * A Java application for digitizing points off of scanned images. This program
 * can manually digitize points off of any image for a wide range of purposes.
 * It can also semi-automatically digitize points off of lines that trend mostly
 * left-to-right and that are functional in nature (one Y value for each X
 * value, no looping back or vertical lines).
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt,    Date: December 15, 2000
 * @version December 24, 2023
 */
public class PlotDigitizer extends MDIApplication implements MainApp {

    /**
     * The resource bundle containing the string resources for this application.
     * Resource bundles are useful for localizing applications. New localities
     * (languages) can be added by adding properties files only -- no code
     * modifications.
     */
    public static final ResourceBundle RESBUNDLE = ResourceBundle.getBundle("appStrings", Locale.getDefault());

    private static final Logger LOGGER = Logger.getLogger(PlotDigitizer.class.getName());

    //  A file filter that recognizes MacOS file types.
    private static final ExtFilenameFilter fnFilter = new ExtFilenameFilter();
    private static final ExtFilenameFilter xmlFilter = new ExtFilenameFilter(".xml", "*.xml");
    private static final ExtFilenameFilter csvFilter = new ExtFilenameFilter(".csv", "*.csv");

    private final AppWindow appWindow;

    /**
     * Main method for this program. This is where the program starts.
     */
    public static void main(String[] arguments) {

        // Schedule a job for the event-dispatching thread:
        // creating and showing this application's GUI.
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                try {

                    // Set the system look and feel to hide that hideous Java LAF.
                    AppUtilities.setSystemLAF();

                    // MacOS users like the menu bar to be at the top of the screen.
                    // This does nothing on non-MacOS platforms.
                    System.setProperty("apple.laf.useScreenMenuBar", "true");

                    // Create an instance of our application to get things rolling.
                    new PlotDigitizer();

                } catch (NoSuchMethodException e) {
                    LOGGER.severe("Unexpected Exception initializing application");
                    AppUtilities.showException(null, RESBUNDLE.getString("unexpectedTitle"),
                            RESBUNDLE.getString("unexpectedMsg"), e);
                    System.exit(0);

                } catch (Exception e) {
                    LOGGER.severe("Unexpected Exception initializing application");
                    AppUtilities.showException(null, "Unexpected Error",
                            "Copy the following message and e-mail it to the author:", e);
                    System.exit(0);
                }
            }
        });

    }

    /**
     * Return a reference to the Logger used by this program.
     */
    public static Logger getLogger() {
        return LOGGER;
    }

    /**
     * Constructs a PlotDigitizer application and open an empty window.
     */
    public PlotDigitizer() throws NoSuchMethodException {
        super(RESBUNDLE.getString("appName"));

        //  Store the resource bundle.
        setResourceBundle(RESBUNDLE);

        //  Set up the application preferences.
        Preferences prefs = new AppPreferences();
        PlotDigitizer.setPreferences(prefs);

        // Set up our filename filter.
        fnFilter.addExtension("gif");
        fnFilter.addExtension("jpg");
        fnFilter.addExtension("jpeg");
        fnFilter.addExtension("png");

        //  Create the menu bar.
        JMenuBar menuBar = createMenuBar();

        // Add the menu bar to this application.
        //  Cause the application to stay open when the last window is closed under MacOS.
        //  When the last window is closed, a menu bar will remain for the application with no
        //  window open.  This is standard MacOS behavior and will not happen on other platforms.
        setFramelessJMenuBar(menuBar);

        // Add an open files handler for MacOS.
        MDIApplication.setOpenFileHandler(new java.awt.desktop.OpenFilesHandler() {
            @Override
            public void openFiles(OpenFilesEvent e) {
                //  Loop over the files to be opened and open each in turn.
                List<File> files = e.getFiles();
                for (File file : files) {
                    openFile(file);
                }
            }
        });
        
        //  Create a main window frame for this application.
        appWindow = new AppWindow();
        appWindow.setVisible(true);
        PlotDigitizer.addWindow(appWindow);
    }

    /**
     * Allows two way communication between the TableWindow and the AppWindow
     *
     * @return Hook to the AppWindow instance to operate on data there
     */
    public AppWindow getAppWindow() {
        return appWindow;
    }

    //-----------------------------------------------------------------------------------
    /**
     * Handle the user requesting a new document window. Creates a new instance
     * of this application's main document window.
     *
     * @param event The event that caused this method to be called. May be
     * <code>null</code> if this method is called by MDIApplication or one of
     * it's subclasses.
     */
    @Override
    public Frame handleNew(ActionEvent event) {
        Frame window = null;
        try {

            window = new AppWindow();
            window.setVisible(true);
            PlotDigitizer.addWindow(window);

        } catch (NoSuchMethodException e) {
            LOGGER.log(Level.SEVERE, "Exception thrown in PlotDigitizer.handleNew() \n", e);
            AppUtilities.showException(null, getResourceBundle().getString("unexpectedTitle"),
                    getResourceBundle().getString("unexpectedMsg"), e);
        }

        return window;
    }

    /**
     * Handles the user choosing "Open" from the File menu. Allows the user to
     * choose an image file to open.
     */
    public void handleOpen(ActionEvent event) {

        try {
            String dir = getPreferences().getLastPath();
            File theFile = AppUtilities.selectFile(null, FileDialog.LOAD, getResourceBundle().getString("fileDialogLoad"),
                    dir, null, getFilenameFilter());
            if (isNull(theFile))
                return;     //  User canceled.

            //  Go and open the file.
            openFile(theFile);

        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, "Exception thrown in PlotDigitizer.handleOpen() \n", e);
            AppUtilities.showException(null, getResourceBundle().getString("unexpectedTitle"),
                    getResourceBundle().getString("unexpectedMsg"), e);
        }

    }

    /**
     * Open the specified file in either the top window (if it is the right kind) or a new window.
     * 
     * @param theFile The file to be opened and read in.
     */
    private void openFile(File theFile) {
        try {
            Window tmp = getTopWindow();
            if (tmp instanceof AppWindow) {
                AppWindow topWindow = (AppWindow)tmp;

                //  If the top window is empty, use it.
                if (topWindow.isEmpty()) {
                    //	Read in the image file and display it.
                    topWindow.readImageFile(theFile);
                    return;
                }
            }
            
            //  Create a new window and load the file into that.
            AppWindow.newWindowFromDataFile(null, theFile);
                        
        } catch (Exception e) {
            e.printStackTrace();
            AppUtilities.showException(null, MDIApplication.getInstance().getResourceBundle().getString("unexpectedTitle"),
                    MDIApplication.getInstance().getResourceBundle().getString("unexpectedMsg"), e);
        }
         
    }
    
    /**
     * Return a reference to this program's GUI application or <code>null</code>
     * if we are running in batch mode.
     */
    @Override
    public final MDIApplication getGUIApplication() {
        return this;
    }

    /**
     * Return a reference to this application's file name filter.
     */
    public static FilenameFilter getFilenameFilter() {
        return fnFilter;
    }

    public static FilenameFilter getXmlFilter() {
        return xmlFilter;
    }

    public static FilenameFilter getCSVFilter() {
        return csvFilter;
    }

    /**
     * Create an about handler for use in this application.
     */
    public AboutHandler createAboutHandler() {

        AboutHandler about = new AboutHandler() {
            @Override
            public void handleAbout​(AboutEvent e) {
                ResourceBundle resBundle = getResourceBundle();

                //  Load in the application's icon image.
                Icon appIcon = getApplicationIcon();
                if (isNull(appIcon))
                    //  Use a generic icon since this app can't read it's custom icon.
                    appIcon = UIManager.getIcon("OptionPane.informationIcon");

                //  Read in the about box text information.
                String credits = readAboutText(resBundle.getString("aboutTextURLStr"));

                String aName = resBundle.getString("appName");
                String aVersion = resBundle.getString("appVersion");
                String aModDate = resBundle.getString("appModDate");
                String copyright = resBundle.getString("copyright");
                StandardMacAboutFrame f = new StandardMacAboutFrame(aName, aVersion + " - " + aModDate);
                f.setApplicationIcon(appIcon);
                f.setCopyright(copyright);
                f.setCredits(credits, "text/html");
                f.setHyperlinkListener(new HyperlinkListener() {
                    @Override
                    public void hyperlinkUpdate(HyperlinkEvent e) {
                        try {
                            MDIApplication.browse(e.getURL().toURI());
                        } catch (Exception err) {
                            LOGGER.log(Level.SEVERE, null, err);
                        }
                    }
                });
                f.setVisible(true);
            }
        };

        return about;
    }

   /**
     * Return the application icon for this program or null if one can not be found.
     * 
     * @return The application icon for this program.
     */
    public Icon getApplicationIcon() {
        ResourceBundle resBundle = getResourceBundle();
        
        //  Load in the application's icon image.
        Icon appIcon = null;
        try {
            URL imgURL = ClassLoader.getSystemResource(resBundle.getString("applicationIconURL"));
            appIcon = new ImageIcon(imgURL);
        } catch (Exception ex) {
            LOGGER.log(Level.WARNING, null, ex);
        }
        
        return appIcon;
    }

    /**
     * Method that handles reading in the contents of the text region of the
     * about box from a text file.
     */
    private String readAboutText(String urlStr) {
        String text;
        try {
            URL url = ClassLoader.getSystemResource(urlStr);

            InputStreamReader reader = new InputStreamReader(url.openStream(), "UTF-8");
            StringWriter writer = new StringWriter();
            int character = reader.read();
            while (character != -1) {
                writer.write(character);
                character = reader.read();
            }
            text = writer.toString();

        } catch (IOException e) {
            LOGGER.log(Level.SEVERE, "Error retrieving About text", e);

            //  Create a standard set of credits if we didn't find the text file.
            text = "<html><body><b>Author:</b><br>Joseph A. Huwaldt<br>";
            text += "<a href=\"mailto:jhuwaldt@users.sourceforge.net\">jhuwaldt@users.sourceforge.net</a><br>";
            text += "<P ALIGN=CENTER><BR>" + getResourceBundle().getString("appName") + " comes with ABSOLUTELY NO WARRANTY;";
            text += "</body></html>";
        }

        return text;
    }

    /**
     * Initializes and displays the menus associated with this window.
     */
    private JMenuBar createMenuBar() throws NoSuchMethodException {
        ResourceBundle resBundle = getResourceBundle();

        //  Create a menu bar.
        JMenuBar menuBar = new JMenuBar();

        // Set up the file menu.
        List<String[]> menuStrings = new ArrayList();
        String[] row = new String[3];
        row[0] = resBundle.getString("newItemText");
        row[1] = resBundle.getString("newItemKey");
        row[2] = "handleNew";
        menuStrings.add(row);
        row = new String[3];
        row[0] = resBundle.getString("openItemText");
        row[1] = resBundle.getString("openItemKey");
        row[2] = "handleOpen";
        menuStrings.add(row);
        menuStrings.add(new String[3]);  //  Blank line
        row = new String[3];
        row[0] = resBundle.getString("closeItemText");
        row[1] = resBundle.getString("closeItemKey");
        row[2] = "handleClose";
        menuStrings.add(row);
        row[0] = resBundle.getString("saveItemText");
        row[1] = resBundle.getString("saveItemKey");
        row[2] = null;
        menuStrings.add(row);
        row[0] = resBundle.getString("saveAsItemText");
        row[1] = null;
        row[2] = null;
        menuStrings.add(row);

        JMenu menu = AppUtilities.buildMenu(this, resBundle.getString("fileMenuText"), menuStrings);
        menuBar.add(menu);

        //  Add a Quit menu item (if it isn't automatically present).
        if (!QuitJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(this.getQuitJMenuItem());
        }
        
        //  Add a Quit handler to ask the user if they want to save unsaved changes before quitting.
        setQuitHandler(new QuitHandler() {
            @Override
            public void handleQuitRequestWith(QuitEvent e, QuitResponse response) {
                handleQuit(e, response);
            }
        });
        
        
        // Set up the edit menu.
        menuStrings.clear();
        row = new String[3];
        row[0] = resBundle.getString("undoItemText");
        row[1] = resBundle.getString("undoItemKey");
        row[2] = null;
        menuStrings.add(row);
        row = new String[3];
        row[0] = resBundle.getString("redoItemText");
        row[1] = null;
        row[2] = null;
        menuStrings.add(row);
        menuStrings.add(new String[3]);  //  Blank line
        row = new String[3];
        row[0] = resBundle.getString("cutItemText");
        row[1] = resBundle.getString("cutItemKey");
        row[2] = null;
        menuStrings.add(row);
        row = new String[3];
        row[0] = resBundle.getString("copyItemText");
        row[1] = resBundle.getString("copyItemKey");
        row[2] = null;
        menuStrings.add(row);
        row = new String[3];
        row[0] = resBundle.getString("pasteItemText");
        row[1] = resBundle.getString("pasteItemKey");
        row[2] = null;
        menuStrings.add(row);

        menu = AppUtilities.buildMenu(this, resBundle.getString("editMenuText"), menuStrings);
        menuBar.add(menu);

         //  Register a Preferences menu item handler.
        MDIApplication.setPreferencesHandler(createPreferencesHandler());
        
        //  Create a Preferences menu item (if not automatically present).
        if (!PreferencesJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(this.getPreferencesJMenuItem());
        }

       //  Add a Preferences menu item.
        PreferencesJMenuItem preferences = this.getPreferencesJMenuItem();
        preferences.setText(resBundle.getString("preferencesItemText"));
        preferences.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                getPreferences().showPreferenceDialog();
            }
        });
        if (!PreferencesJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(preferences);
        }

        //	Register an about menu item handler.
        setAboutHandler(createAboutHandler());
        
        //  Create an about menu item (if not automatically present).
        if (!AboutJMenuItem.isAutomaticallyPresent()) {
            //  Create Help menu.
            menu = new JMenu(resBundle.getString("helpMenuText"));
            menuBar.add(menu);
            
            //  Add the "About" item to the Help menu.
            menu.add(this.getAboutJMenuItem());
        }
        
        return menuBar;
    }

    /**
     * Return a preferences handler that displays the preference dialog for this application.
     * 
     * @return A preferences handler
     */
    public PreferencesHandler createPreferencesHandler() {
        PreferencesHandler handler = new PreferencesHandler() {
            @Override
            public void handlePreferences(PreferencesEvent e) {
                PlotDigitizer.this.getPreferences().showPreferenceDialog();
            }
        };
        return handler;
    }
    
}
